//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#pragma package(smart_init)
#include "OpenGl.h"

#ifdef USESGI
#pragma link "opengl.lib"
#endif


//---------------------------------------------------------------------------
// ValidCtrCheck is used to assure that the components created do not have
// any pure virtual functions.
//
static GLfloat vertices[] =
    {
    -0.5, -0.5, -0.5,
    0.5, -0.5, -0.5,
    0.5, 0.5, -0.5,
    -0.5, 0.5, -0.5,
    -0.5, -0.5, 0.5,
    0.5, -0.5, 0.5,
    0.5, 0.5, 0.5,
    -0.5, 0.5, 0.5
    };


static inline void ValidCtrCheck(TOpenGl *)
{
    new TOpenGl(NULL);
}
//---------------------------------------------------------------------------
__fastcall TOpenGl::~TOpenGl()
{
    delete FEyePosition;
    delete FEyeTarget;
    delete FEyeUp;
    for (int i=0; i<MAXLIGHTS; i++)
        {
        delete luci[i];
        }
    delete FCanvas;

    for (int i=0; i<MAXBITMAPS; i++)
        delete pictures[i];

    for (int i=0; i<parts->Count; i++)
        {
        TPart *pP = (TPart *)parts->Items[i];
        delete pP;
        }
}

__fastcall TOpenGl::TOpenGl(TComponent* Owner)
    : TWinControl(Owner)
{
    FSuggestedSelection = 0;
    FEyePosition = new T3DVector();
    FEyePosition->Z = -8.0;
    FEyeTarget = new T3DVector();
    FEyeUp = new T3DVector();
    FEyeUp->Y = 1.0;

    for (int i=0; i<MAXLIGHTS; i++)
        {
        luci[i] = new TLuce();
        luci[i]->OnChange = SetLuci;
        }


    for (int i=0; i<MAXBITMAPS; i++)
        textureIndexes[i] = 0;
  FFps = 0.0;

  FProjectionOrtho = false;
  FProjectionWidth = 1.0;
  FProjectionHeight = 1.0;
  FProjectionNear = 1.0;
  FProjectionFar = 10.0;
  FProjectionYAngle = 90.0;

  FAnimation = false;
  LastPerfCounter.QuadPart = 0;
  QueryPerformanceFrequency(&PerfFreq);
  FCanvas = new TControlCanvas();
  FCanvas->Control = this;

  for (int i=0; i<MAXBITMAPS; i++)
    pictures[i] = new TPicture();

  parts = new TList();

#ifdef NOOO
  TPart * pp = (TPart *)new TPartCube();
  pp->PX = -2.0;
  pp->PZ = -4.0;
  pp->OnChange = Refresh;
  AddObject(0, pp);
  pp = (TPart *)new TPartSphere();
  pp->PX = 2.0;
  pp->PZ = -4.0;
  pp->OnChange = Refresh;
  AddObject(0, pp);
#endif  
}

void __fastcall TOpenGl::AddObject(TPart * pParent, TPart * pObj)
{
    TPartGroup * pGroup = dynamic_cast<TPartGroup *>(pParent);
    if (pGroup)
        {
        pGroup->AddObject(pObj);
        }
    else
        parts->Add(pObj);
}

void __fastcall TOpenGl::DeleteObject(TPart * pObj)
{
  if (!pObj)
    return;
  TPartGroup * pGroup = pObj->Parent();
  if (pGroup)
    pGroup->DeleteObject(pObj);
  else
    {
    int idx = parts->IndexOf(pObj);
    if (idx >= 0)
        parts->Delete(idx);
    }
  FSuggestedSelection = 0;
  if (FOnListChanged)
    FOnListChanged(this);
  Refresh(this);
}


//---------------------------------------------------------------------------
namespace Opengl
{
    void __fastcall PACKAGE Register()
    {
        TComponentClass classes[5] =
            {
            __classid(TOpenGl),
            __classid(TFloatEdit),
            __classid(TPos3D),
            __classid(TColorEdit),
            __classid(TAngleEdit)
            };
        RegisterComponents("3D", classes, 4);
    }
}
//---------------------------------------------------------------------------
void __fastcall TOpenGl::CreateParams(TCreateParams &Params)
{
    TWinControl::CreateParams(Params);
    Params.WindowClass.style |= (CS_OWNDC | CS_HREDRAW | CS_VREDRAW);
}

void TOpenGl::SetProjection(void)
{
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
//    if (OnSetProjection)
//        OnSetProjection(this);
    if (FProjectionOrtho)
        glOrtho(-FProjectionWidth/2, FProjectionWidth/2, -FProjectionHeight/2, FProjectionHeight/2, FProjectionNear, FProjectionFar);
    else
        {
        GLdouble aspect = Width;
        aspect /= Height;
//        glFrustum(-FProjectionWidth/2, FProjectionWidth/2, -FProjectionHeight/2, FProjectionHeight/2, FProjectionNear, FProjectionFar);
        gluPerspective( FProjectionYAngle,
		       aspect,
		       FProjectionNear,
		       FProjectionFar);
        }
    gluLookAt(  FEyePosition->Element[0], FEyePosition->Element[1], FEyePosition->Element[2],
                FEyeTarget->Element[0], FEyeTarget->Element[1], FEyeTarget->Element[2],
                FEyeUp->Element[0], FEyeUp->Element[1], FEyeUp->Element[2]);
    glMatrixMode(GL_MODELVIEW);
}

void TOpenGl::defaultInitRC(void)
{
    SetProjection();
    /* set viewing projection */
    glTranslatef(0.0F, 0.0F, -2.0F);

    /* position object */
    glRotatef(30.0F, 1.0F, 0.0F, 0.0F);
    glRotatef(30.0F, 0.0F, 1.0F, 0.0F);

    glEnable(GL_DEPTH_TEST);
//    glEnable(GL_CULL_FACE);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glCullFace(GL_BACK);

   glEnableClientState(GL_VERTEX_ARRAY);
   glVertexPointer(3, GL_FLOAT, 0, vertices);
}

void TOpenGl::setupPixelFormat()
{
    PIXELFORMATDESCRIPTOR pfd = {
        sizeof(PIXELFORMATDESCRIPTOR),  /* size */
        1,                              /* version */
        PFD_SUPPORT_OPENGL |
        PFD_DRAW_TO_WINDOW |
        PFD_DOUBLEBUFFER,               /* support double-buffering */
        PFD_TYPE_RGBA,                  /* color type */
        16,                             /* prefered color depth */
        0, 0, 0, 0, 0, 0,               /* color bits (ignored) */
        0,                              /* no alpha buffer */
        0,                              /* alpha bits (ignored) */
        0,                              /* no accumulation buffer */
        0, 0, 0, 0,                     /* accum bits (ignored) */
        16,                             /* depth buffer */
        0,                              /* no stencil buffer */
        0,                              /* no auxiliary buffers */
        PFD_MAIN_PLANE,                 /* main layer */
        0,                              /* reserved */
        0, 0, 0,                        /* no layer, visible, damage masks */
    };
    int pixelFormat;

    pixelFormat = ChoosePixelFormat(Canvas->Handle, &pfd);
    if (pixelFormat == 0) {
        MessageBox(Handle, "ChoosePixelFormat failed.", "Error",
                MB_ICONERROR | MB_OK);
    }

    if (SetPixelFormat(Canvas->Handle, pixelFormat, &pfd) != TRUE) {
        MessageBox(Handle, "SetPixelFormat failed.", "Error",
                MB_ICONERROR | MB_OK);
    }
}

void __fastcall TOpenGl::Loaded()
{
    setupPixelFormat();
    hGLRC = wglCreateContext(Canvas->Handle);
    wglMakeCurrent(Canvas->Handle, hGLRC);
    defaultInitRC();
    if (FOnInitRC)
        FOnInitRC(this);
}

void __fastcall TOpenGl::SetAnimation(bool AAnimation)
{
    if (AAnimation == FAnimation)
        return;
    FAnimation = AAnimation;
    InvalidateRect(Handle, NULL, false);
}

void __fastcall TOpenGl::Paint(void)
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glMatrixMode(GL_MODELVIEW);

    glLoadIdentity();

    if (FAnimation && FOnSetProjection)
        {
        OnSetProjection(this);
        }
    if (OnPaint)
        OnPaint(this);
    else
        {
        for (int i=0; i<parts->Count; i++)
            {
            glPushMatrix();
            TPart *pP = (TPart *)parts->Items[i];
            pP->Paint();
            glPopMatrix();
            }
        }

    SwapBuffers(Canvas->Handle);

    if (FAnimation && FOnFrame)
        {
        FOnFrame(this);
        InvalidateRect(Handle, NULL, false);
        }

    LARGE_INTEGER pc;
    QueryPerformanceCounter(&pc);
    FFps = PerfFreq.QuadPart;
    FFps /= (pc.QuadPart-LastPerfCounter.QuadPart);
    LastPerfCounter = pc;
}

void __fastcall TOpenGl::PaintWindow(HDC DC)
{
  FCanvas->Lock();
  try
    {
    FCanvas->Handle = DC;
    try
        {
        Paint();
        }
    __finally
        {
        FCanvas->Handle = 0;
        }
    }
  __finally
    {
    FCanvas->Unlock();
    }
}

void __fastcall TOpenGl::WMPaint(TWMPaint Message)
{
  TPaintStruct PS;
  HDC DC = Message.DC;
  if (DC == 0)
     {
     DC = BeginPaint(Handle, &PS);
     PaintWindow(DC);
     EndPaint(Handle, &PS);
     }
  else
     PaintWindow(DC);
}

void __fastcall TOpenGl::WMSize(TWMSize Message)
{
    TRect Rect;
    GetWindowRect(Handle, (tagRECT *) &Rect);
    UpdateBoundsRect(Rect);
    glViewport(0, 0, Width, Height);
}

void __fastcall TOpenGl::SetProjectionOrtho(bool AProjectionOrtho)
{
    FProjectionOrtho = AProjectionOrtho;
    SetProjection();
    InvalidateRect(Handle, NULL, false);
}

void __fastcall TOpenGl::SetProjectionWidth(float AProjectionWidth)
{
    FProjectionWidth = AProjectionWidth;
    SetProjection();
    InvalidateRect(Handle, NULL, false);
}

void __fastcall TOpenGl::SetProjectionHeight(float AProjectionHeight)
{
    FProjectionHeight = AProjectionHeight;
    SetProjection();
    InvalidateRect(Handle, NULL, false);
}

void __fastcall TOpenGl::SetProjectionNear(float AProjectionNear)
{
    FProjectionNear = AProjectionNear;
    SetProjection();
    InvalidateRect(Handle, NULL, false);
}

void __fastcall TOpenGl::SetProjectionFar(float AProjectionFar)
{
    FProjectionFar = AProjectionFar;
    SetProjection();
    InvalidateRect(Handle, NULL, false);
}

void __fastcall TOpenGl::SetProjectionYAngle(float AProjectionYAngle)
{
    FProjectionYAngle = AProjectionYAngle;
    SetProjection();
    InvalidateRect(Handle, NULL, false);
}

void __fastcall TOpenGl::Refresh(TObject *Sender)
{
    InvalidateRect(Handle, NULL, false);
}

int __fastcall TOpenGl::UseTexture(int index)
{
    glCallList(textureIndexes[index]);
    return textureIndexes[index];
}

int __fastcall TOpenGl::SetTexture(int index, Graphics::TBitmap * pBitmap)
{
    if (textureIndexes[index])
        glDeleteLists(textureIndexes[index],1);
    textureIndexes[index] = glGenLists(1);
    glNewList(textureIndexes[index], GL_COMPILE);
    BYTE * ima= new BYTE[pBitmap->Width*pBitmap->Height*3];
//    TPixelFormat pf = pBitmap->PixelFormat;
    for (int y = 0; y < pBitmap->Height; y++)
    {
      for (int x = 0; x < pBitmap->Width; x++)
      {
      union
        {
        TColor c;
        struct
            {
            BYTE red;
            BYTE green;
            BYTE blue;
            BYTE x;
            } b;
        } co;
       co.c = pBitmap->Canvas->Pixels[x][y];
       ima[y*pBitmap->Width*3+x*3] = co.b.red;
       ima[y*pBitmap->Width*3+x*3+1] = co.b.green;
       ima[y*pBitmap->Width*3+x*3+2] = co.b.blue;
      }
    }

    gluBuild2DMipmaps(GL_TEXTURE_2D, 3,
                pBitmap->Width,
                pBitmap->Height,
		        GL_RGB, GL_UNSIGNED_BYTE, ima);
    glEndList();
    delete ima;

    return textureIndexes[index];
}

void  __fastcall TOpenGl::SetPicture(int n, TPicture * Apicture)
{
    pictures[n]->Assign(Apicture);
    SetTexture(n, pictures[n]->Bitmap);
}

void __fastcall TOpenGl::SetEyePosition(T3DVector * Value)
{
    FEyePosition->Assign((TPersistent *)Value);
    SetProjection();
    InvalidateRect(Handle, NULL, false);
}

void __fastcall TOpenGl::SetEyeTarget(T3DVector * Value)
{
    FEyeTarget->Assign((TPersistent *)Value);
    SetProjection();
    InvalidateRect(Handle, NULL, false);
}

void __fastcall TOpenGl::SetEyeUp(T3DVector * Value)
{
    FEyeUp->Assign((TPersistent *)Value);
    SetProjection();
    InvalidateRect(Handle, NULL, false);
}

TPart * __fastcall TOpenGl::AddCube(TPart * pParent, const char * name)
{
  TPart * pp = (TPart *)new TPartCube();
  pp->OnChange = Refresh;
  if (name)
    pp->Name = name;
  AddObject(pParent, pp);
  FSuggestedSelection = pp;
  if (FOnListChanged)
    FOnListChanged(this);
  Refresh(this);
  return pp;
}

TPart * __fastcall TOpenGl::AddQuad(TPart * pParent, const char * name)
{
  TPart * pp = (TPart *)new TPartQuad();
  pp->OnChange = Refresh;
  if (name)
    pp->Name = name;
  AddObject(pParent, pp);
  FSuggestedSelection = pp;
  if (FOnListChanged)
    FOnListChanged(this);
  Refresh(this);
  return pp;
}

TPart * __fastcall TOpenGl::Paste(TPart * pParent, TPart * pp)
{
  pp->OnChange = Refresh;
  AddObject(pParent, pp);
  FSuggestedSelection = pp;
  if (FOnListChanged)
    FOnListChanged(this);
  Refresh(this);
  return pp;
}

TPart * __fastcall TOpenGl::AddGroup(TPart * pParent, const char * name)
{
  TPart * pp = (TPart *)new TPartGroup();
  pp->OnChange = Refresh;
  if (name)
    pp->Name = name;
  AddObject(pParent, pp);
  FSuggestedSelection = pp;
  if (FOnListChanged)
    FOnListChanged(this);
  Refresh(this);
  return pp;
}

TPart * __fastcall TOpenGl::AddSphere(TPart * pParent, const char * name)
{
  TPart * pp = (TPart *)new TPartSphere();
  pp->OnChange = Refresh;
  if (name)
    pp->Name = name;
  AddObject(pParent, pp);
  FSuggestedSelection = pp;
  if (FOnListChanged)
    FOnListChanged(this);
  Refresh(this);
  return pp;
}

TPart * __fastcall TOpenGl::AddCylinder(TPart * pParent, const char * name)
{
  TPart * pp = (TPart *)new TPartCylinder();
  pp->OnChange = Refresh;
  if (name)
    pp->Name = name;
  AddObject(pParent, pp);
  FSuggestedSelection = pp;
  if (FOnListChanged)
    FOnListChanged(this);
  Refresh(this);
  return pp;
}

TPart * __fastcall TOpenGl::AddDisk(TPart * pParent, const char * name)
{
  TPart * pp = (TPart *)new TPartDisk();
  pp->OnChange = Refresh;
  if (name)
    pp->Name = name;
  AddObject(pParent, pp);
  FSuggestedSelection = pp;
  if (FOnListChanged)
    FOnListChanged(this);
  Refresh(this);
  return pp;
}

void __fastcall TOpenGl::SetLuci(TObject *Sender)
{
    glPushMatrix();
    glLoadIdentity();
    for (int i=0; i<MAXLIGHTS; i++)
        {
        if (luci[i]->Enabled)
            {
            glEnable(GL_LIGHT0+i);
            glLightfv(GL_LIGHT0+i,
                GL_POSITION,
                luci[i]->Position->Vector());
            glLightfv(GL_LIGHT0+i,
                GL_AMBIENT,
                luci[i]->Ambient->Vector());
            glLightfv(GL_LIGHT0+i,
                GL_DIFFUSE,
                luci[i]->Diffuse->Vector());
            glLightfv(GL_LIGHT0+i,
                GL_SPECULAR,
                luci[i]->Specular->Vector());
            }
        else
            {
            glDisable(GL_LIGHT0+i);
            }
        }
    Refresh(this);
    glPopMatrix();
}

void __fastcall TOpenGl::SetLight(int n, TLuce * Value)
{
    if (n<0 || n>=MAXLIGHTS)
        return;
    luci[n]->Assign(Value);
}

TLuce * __fastcall TOpenGl::GetLight(int n)
{
    if (n<0 || n>=MAXLIGHTS)
        return 0;
    return luci[n];
}


TLuce::TLuce()
{
        Position = new T3DVector();
        Enabled = false;
        Ambient = new TColorVector();
        Ambient->R = 1.0;
        Ambient->G = 1.0;
        Ambient->B = 1.0;
        Ambient->A = 1.0;
        Diffuse = new TColorVector();
        Diffuse->R = 1.0;
        Diffuse->G = 1.0;
        Diffuse->B = 1.0;
        Diffuse->A = 1.0;
        Specular = new TColorVector();
        Specular->R = 1.0;
        Specular->G = 1.0;
        Specular->B = 1.0;
        Specular->A = 1.0;
        Position->OnChange = Changed;
        Ambient->OnChange = Changed;
        Diffuse->OnChange = Changed;
        Specular->OnChange = Changed;
}

void __fastcall TLuce::Assign(TPersistent * Value)
{
    TLuce * pV= (TLuce *)Value;
    Position->Assign(pV->Position);
    Enabled = pV->Enabled;
    Ambient->Assign(pV->Ambient);
    Diffuse->Assign(pV->Diffuse);
    Specular->Assign(pV->Specular);
}

__fastcall TLuce::~TLuce()
{
    if (Position)
        delete Position;
    if (Ambient)
        delete Ambient;
    if (Diffuse)
        delete Diffuse;
    if (Specular)
        delete Specular;
}

bool TLuce::SaveTo(TSceneFile & sf)
{
    sf.writeInt(FEnabled ? 1 : 0);
    Position->SaveTo(sf);
    Ambient->SaveTo(sf);
    Diffuse->SaveTo(sf);
    Specular->SaveTo(sf);
    return sf.OK();
}

bool TLuce::LoadFrom(TSceneFile & sf)
{
    int tmp;
    sf.readInt(VERSION_0, tmp, 0);
    FEnabled = (tmp==1);
    Position->LoadFrom(VERSION_0, sf);
    Ambient->LoadFrom(VERSION_0, sf, 1.0, 1.0, 1.0, 1.0);
    Diffuse->LoadFrom(VERSION_0, sf, 1.0, 1.0, 1.0, 1.0);
    Specular->LoadFrom(VERSION_0, sf, 1.0, 1.0, 1.0, 1.0);
    return sf.OK();
}

bool TOpenGl::SaveTo(String fileName)
{
    TSceneFile sf(fileName.c_str(), true, VERSION_0);

    sf.writeInt(FAnimation ? 1 : 0);
    sf.writeInt(FProjectionOrtho ? 1 : 0);

    sf.writeFloat(FProjectionWidth);
    sf.writeFloat(FProjectionHeight);
    sf.writeFloat(FProjectionNear);
    sf.writeFloat(FProjectionFar);
    sf.writeFloat(FProjectionYAngle);

    FEyePosition->SaveTo(sf);
    FEyeTarget->SaveTo(sf);
    FEyeUp->SaveTo(sf);

    sf.writeInt(parts->Count);

    for (int i=0; i<parts->Count; i++)
        {
        TPart *pP = (TPart *)parts->Items[i];
        pP->SaveTo(sf);
        }

    for (int i=0; i<MAXLIGHTS; i++)
        {
        luci[i]->SaveTo(sf);
        }
    return sf.OK();
}

bool TOpenGl::LoadFrom(String fileName)
{
    TSceneFile sf(fileName.c_str(), false,0);

    int tmp;

    sf.readInt(VERSION_0, tmp, 0);
    FAnimation = (tmp==1);
    sf.readInt(VERSION_0, tmp, 0);
    FProjectionOrtho = (tmp==1);

    sf.readFloat(VERSION_0, FProjectionWidth, 1.0);
    sf.readFloat(VERSION_0, FProjectionHeight, 1.0);
    sf.readFloat(VERSION_0, FProjectionNear, 1.0);
    sf.readFloat(VERSION_0, FProjectionFar, 10.0);
    sf.readFloat(VERSION_0, FProjectionYAngle, 90.0);

    FEyePosition->LoadFrom(VERSION_0, sf);
    FEyeTarget->LoadFrom(VERSION_0, sf);
    FEyeUp->LoadFrom(VERSION_0, sf);

    int c=0;
    parts->Clear();
    sf.readInt(VERSION_0, c, 0);
    for (int i=0; i<c; i++)
        {
        TPart * pP = TPart::Load(sf);
        if (pP)
            {
            pP->OnChange = Refresh;
            parts->Add(pP);
            }
        }

    for (int i=0; i<MAXLIGHTS; i++)
        {
        luci[i]->LoadFrom(sf);
        }

    SetLuci(this);

    if (OnListChanged)
        OnListChanged(this);
    return sf.OK();
}

